from __future__ import annotations

import warnings
from typing import Optional, Union, Tuple

from selenium.webdriver.remote.switch_to import SwitchTo
from selenium.webdriver.remote.webdriver import WebDriver

from selene.common.helpers import to_by
from selene.core._actions import _Actions
from selene.core.configuration import Config
from selene.core.entity import WaitingEntity, Element, Collection
from selene.core.locator import Locator
from selene.support.webdriver import WebHelper


class Browser(WaitingEntity['Browser']):
    def __init__(self, config: Optional[Config] = None):
        config = Config() if config is None else config
        super().__init__(config)

    def with_(self, config: Optional[Config] = None, **config_as_kwargs) -> Browser:
        return (
            Browser(config)
            if config
            else Browser(self.config.with_(**config_as_kwargs))
        )

    def __str__(self):
        return 'browser'

    @property
    def driver(self) -> WebDriver:
        return self.config.driver

    # TODO: consider making it callable (self.__call__() to be shortcut to self.__raw__ ...)

    @property
    def __raw__(self):
        return self.config.driver

    # @property
    # def actions(self) -> ActionChains:
    #     """
    #     It's kind of just a shortcut for pretty low level actions from selenium webdriver
    #     Yet unsure about this property here:)
    #     comparing to usual high level Selene API...
    #     Maybe later it would be better to make own Actions with selene-style retries, etc.
    #     """
    #     return ActionChains(self.config.driver)

    @property
    def _actions(self) -> _Actions:
        return _Actions(self.config)

    # --- Element builders --- #

    # TODO: consider None by default,
    #       and *args, **kwargs to be able to pass custom things
    #       to be processed by config.location_strategy
    #       and by default process none as "element to skip all actions on it"
    def element(
        self, css_or_xpath_or_by: Union[str, Tuple[str, str], Locator]
    ) -> Element:
        if isinstance(css_or_xpath_or_by, Locator):
            return Element(css_or_xpath_or_by, self.config)

        by = to_by(css_or_xpath_or_by)

        return Element(
            Locator(f'{self}.element({by})', lambda: self.driver.find_element(*by)),
            self.config,
        )

    def all(
        self, css_or_xpath_or_by: Union[str, Tuple[str, str], Locator]
    ) -> Collection:
        if isinstance(css_or_xpath_or_by, Locator):
            return Collection(css_or_xpath_or_by, self.config)

        by = to_by(css_or_xpath_or_by)

        return Collection(
            Locator(f'{self}.all({by})', lambda: self.driver.find_elements(*by)),
            self.config,
        )

    # --- High Level Commands--- #

    def open(self, relative_or_absolute_url: Optional[str] = None) -> Browser:
        # TODO: should we keep it less pretty but more KISS? like:
        # self.config._driver_get_url_strategy(self.config)(relative_or_absolute_url)
        self.config._executor.get_url(relative_or_absolute_url)

        return self

    def switch_to_next_tab(self) -> Browser:
        from selene.core import query

        self.driver.switch_to.window(query.next_tab(self))

        # TODO: should we use waiting version here (and in other similar cases)?
        # self.perform(Command(
        #     'open next tab',
        #     lambda browser: browser.driver.switch_to.window(query.next_tab(self))))

        return self

    def switch_to_previous_tab(self) -> Browser:
        from selene.core import query

        self.driver.switch_to.window(query.previous_tab(self))
        return self

    def switch_to_tab(self, index_or_name: Union[int, str]) -> Browser:
        if isinstance(index_or_name, int):
            index = index_or_name
            from selene.core import query

            self.driver.switch_to.window(query.tab(index)(self))
        else:
            self.driver.switch_to.window(index_or_name)

        return self

    # TODO: consider deprecating
    @property
    def switch_to(self) -> SwitchTo:
        return self.driver.switch_to

    # TODO: should we add also a shortcut for self.driver.switch_to.alert ?
    #       if we don't need to switch_to.'back' after switch to alert - then for sure we should...
    #       question is - should we implement our own alert as waiting entity?

    def quit(self) -> None:
        """
        Quits the driver.

        If the driver was not even set, will build it just to quit it:D.

        Will fail if the driver was already quit or crashed.
        """
        self.driver.quit()

    # TODO: consider deprecating, it does not close browser, it closes current tab/window
    def close(self) -> Browser:
        self.driver.close()
        return self

    # --- Deprecated --- #

    # TODO: should we keep it?
    def execute_script(self, script, *args):
        warnings.warn(
            'consider using browser.driver.execute_script '
            'instead of browser.execute_script',
            PendingDeprecationWarning,
        )
        return self.driver.execute_script(script, *args)

    # TODO: should we move it to query.* and/or command.*?
    #       like `browser.get(query.screenshot)` ?
    #       like `browser.perform(command.save_screenshot)` ?
    # TODO: deprecate file name, use path
    #       because we can path folder path not file path and it will work
    def save_screenshot(self, file: Optional[str] = None):
        warnings.warn(
            'browser.save_screenshot is deprecated, '
            'use browser.get(query.screenshot_saved())',
            DeprecationWarning,
        )

        from selene.core import query  # type: ignore

        return self.get(query.screenshot_saved())  # type: ignore

    @property
    def last_screenshot(self) -> str:
        warnings.warn(
            'browser.last_screenshot is deprecated, '
            'use browser.config.last_screenshot',
            DeprecationWarning,
        )
        return self.config.last_screenshot  # type: ignore

    # TODO: consider moving this to browser command.save_page_source(filename)
    def save_page_source(self, file: Optional[str] = None) -> Optional[str]:
        warnings.warn(
            'browser.save_page_source is deprecated, '
            'use browser.get(query.page_source_saved())',
            DeprecationWarning,
        )

        if file is None:
            file = self.config._generate_filename(suffix='.html')  # type: ignore

        saved_file = WebHelper(self.driver).save_page_source(file)

        self.config.last_page_source = saved_file  # type: ignore

        return saved_file

    @property
    def last_page_source(self) -> str:
        warnings.warn(
            'browser.last_page_source is deprecated, '
            'use browser.config.last_page_source',
            DeprecationWarning,
        )
        return self.config.last_page_source  # type: ignore

    def close_current_tab(self) -> Browser:
        warnings.warn(
            'deprecated because the «tab» term is not relevant for mobile; '
            'use a `browser.close()` or `browser.driver.close()` instead',
            DeprecationWarning,
        )
        self.driver.close()
        return self

    def clear_local_storage(self) -> Browser:
        warnings.warn(
            'deprecated because of js nature and not-relevance for mobile; '
            'use `browser.perform(command.js.clear_local_storage)` instead',
            DeprecationWarning,
        )
        from selene.core import command

        self.perform(command.js.clear_local_storage)
        return self

    def clear_session_storage(self) -> Browser:
        warnings.warn(
            'deprecated because of js nature and not-relevance for mobile; '
            'use `browser.perform(command.js.clear_session_storage)` instead',
            DeprecationWarning,
        )
        from selene.core import command

        self.perform(command.js.clear_session_storage)
        return self



# MIT License
#
# Copyright (c) 2024 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations
from typing import Optional, List, Union, overload

from selenium.webdriver.common.actions.wheel_input import ScrollOrigin
from selenium.webdriver.remote.webelement import WebElement

from selenium.webdriver import ActionChains
from selenium.webdriver.common.action_chains import AnyDevice

from selene.core.entity import Element
from selene.core.configuration import Config
from selene.core.exceptions import _SeleneError
from selene.core.wait import Command, Query


@overload
def _ensure_located(element: Element | WebElement) -> WebElement: ...


@overload
def _ensure_located(element: Element | WebElement | None) -> WebElement | None: ...


def _ensure_located(element):
    return (
        element.get(Query('locate webelement', lambda it: it.locate()))
        if isinstance(element, Element)
        else element
    )


# TODO: refactor docstrings to the style used in Selene
# TODO: how will it work with Appium?
# TODO: should not we name it ActionsChain to differentiate from w3c actions?
# Selene does not like aliases because of PEP20, but here, in order to provide a better API,
# yet keeping compatibility with Selenium's ActionChains, we are using a lot of aliases
# TODO: but should we?
class _Actions:
    def __init__(
        self,
        config: Config,
        *,
        duration: int = 250,  # TODO: do we need this?
        #       should we have it in config instead?
        devices: Optional[List[AnyDevice]] = None,
    ):
        # TODO: below we pass driver as non lazy instance,
        #       that may lead to some issues in some multi-browser scenarios...
        #       should we bother about them?
        self._config = config
        # TODO: should we check here if driver is appium based
        #       and create proper instance correspondingly?
        self._chain = ActionChains(config.driver, duration, devices)

    def perform(self) -> None:
        def actions(chain: ActionChains):
            """
            if this fn will fail, the final error on wait will be something like:

            Timed out after 1s, while waiting for:
            <selenium.webdriver.common.action_chains.ActionChains object at 0x103115760>
            .{'actions': [[{'type': 'pointerMove', 'duration': 250, 'x': 0, 'y': 0,
             'origin': {'element-6066-11e4-a52e-4f735466cecf':
             '7F810F7CF30DD4907173DF4247841EB2_element_3'}},
             {'type': 'pointerDown', 'duration': 0, 'button': 0},
             {'type': 'pointerUp', 'duration': 0, 'button': 0}],
             [{'type': 'pause', 'duration': 0}, {'type': 'pause', 'duration': 0},
             {'type': 'pause', 'duration': 0}]]}

            TODO: can and should we improve it?
            """
            chain.perform()

        self._config.wait(self._chain).for_(Command(str(self.__encoded), actions))  # type: ignore

    @property
    def __encoded(self):
        return {
            'actions': list(
                filter(
                    None,
                    [
                        device.encode()['actions'] or None
                        for device in self._chain.w3c_actions.devices
                    ],
                )
            )
        }

    def click(self, on_element: Element | WebElement | None = None) -> _Actions:
        """Clicks an element.

        :Args:
         - on_element: The element to click.
           If None, clicks on current mouse position.
        """
        self._chain.click(_ensure_located(on_element))
        return self

    def click_and_hold(
        self, on_element: Element | WebElement | None = None
    ) -> _Actions:
        """Holds down the left mouse button on an element.

        :Args:
         - on_element: The element to mouse down.
           If None, clicks on current mouse position.
        """
        self._chain.click_and_hold(_ensure_located(on_element))
        return self

    def context_click(self, on_element: Element | WebElement | None = None) -> _Actions:
        """Performs a context-click (right click) on an element.

        :Args:
         - on_element: The element to context-click.
           If None, clicks on current mouse position.
        """
        self._chain.context_click(_ensure_located(on_element))
        return self

    def double_click(self, on_element: Element | WebElement | None = None) -> _Actions:
        """Double-clicks an element.

        :Args:
         - on_element: The element to double-click.
           If None, clicks on current mouse position.
        """
        self._chain.double_click(_ensure_located(on_element))
        return self

    def drag_and_drop(
        self, source: Element | WebElement, target: Element | WebElement
    ) -> _Actions:
        """Holds down the left mouse button on the source element, then moves
        to the target element and releases the mouse button.

        :Args:
         - source: The element to mouse down.
         - target: The element to mouse up.
        """
        self._chain.drag_and_drop(_ensure_located(source), _ensure_located(target))
        return self

    def drag_and_drop_by_offset(
        self, source: Element | WebElement, x: int, y: int
    ) -> _Actions:
        """Holds down the left mouse button on the source element, then moves
        to the target offset and releases the mouse button.

        :Args:
         - source: The element to mouse down.
         - xoffset: X offset to move to.
         - yoffset: Y offset to move to.
        """
        self._chain.drag_and_drop_by_offset(_ensure_located(source), x, y)
        return self

    def key_down(
        self, value: str, element: Element | WebElement | None = None
    ) -> _Actions:
        """Sends a key press only, without releasing it. Should only be used
        with modifier keys (Control, Alt and Shift).

        :Args:
         - value: The modifier key to send. Values are defined in `Keys` class.
         - element: The element to send keys.
           If None, sends a key to current focused element.

        Example, pressing ctrl+c::

            ActionChains(driver).key_down(Keys.CONTROL).send_keys('c').key_up(Keys.CONTROL).perform()
        """
        self._chain.key_down(value, _ensure_located(element))
        return self

    def key_up(
        self, value: str, element: Element | WebElement | None = None
    ) -> _Actions:
        """Releases a modifier key.

        :Args:
         - value: The modifier key to send. Values are defined in Keys class.
         - element: The element to send keys.
           If None, sends a key to current focused element.

        Example, pressing ctrl+c::

            ActionChains(driver).key_down(Keys.CONTROL).send_keys('c').key_up(Keys.CONTROL).perform()
        """
        self._chain.key_up(value, _ensure_located(element))
        return self

    def move_by_offset(self, x: int, y: int) -> _Actions:
        """Moving the mouse to an offset from current mouse position.

        :Args:
         - x: X offset to move to, as a positive or negative integer.
         - y: Y offset to move to, as a positive or negative integer.
        """
        self._chain.move_by_offset(x, y)
        return self

    def move_to_element(self, to_element: Element | WebElement) -> _Actions:
        """Moving the mouse to the middle of an element.

        :Args:
         - to_element: The WebElement to move to.
        """
        self._chain.move_to_element(_ensure_located(to_element))
        return self

    def move_to(self, element: Element | WebElement) -> _Actions:
        """Alias for move_to_element"""
        return self.move_to_element(element)

    def move_to_element_with_offset(
        self, to_element: Element | WebElement, xoffset: int, yoffset: int
    ) -> _Actions:
        """Move the mouse by an offset of the specified element. Offsets are
        relative to the in-view center point of the element.

        :Args:
         - to_element: The WebElement to move to.
         - xoffset: X offset to move to, as a positive or negative integer.
         - yoffset: Y offset to move to, as a positive or negative integer.
        """
        self._chain.move_to_element_with_offset(
            _ensure_located(to_element), xoffset, yoffset
        )
        return self

    def move_with_offset_to(
        self, element: Element | WebElement, x: int, y: int
    ) -> _Actions:
        """Alias for move_to_element_with_offset"""
        return self.move_to_element_with_offset(element, x, y)

    def pause(self, seconds: float | int) -> _Actions:
        """Pause all inputs for the specified duration in seconds."""
        self._chain.pause(seconds)
        return self

    def release(self, on_element: Element | WebElement | None = None) -> _Actions:
        """Releasing a held mouse button on an element.

        :Args:
         - on_element: The element to mouse up.
           If None, releases on current mouse position.
        """
        self._chain.release(_ensure_located(on_element))
        return self

    def send_keys(self, *keys_to_send: str) -> _Actions:
        """Sends keys to current focused element.

        :Args:
         - keys_to_send: The keys to send.  Modifier keys constants can be found in the
           'Keys' class.
        """
        self._chain.send_keys(*keys_to_send)
        return self

    def send_keys_to_element(
        self, element: Element | WebElement, *keys_to_send: str
    ) -> _Actions:
        """Sends keys to an element.

        :Args:
         - element: The element to send keys.
         - keys_to_send: The keys to send.  Modifier keys constants can be found in the
           'Keys' class.
        """
        self.click(element)
        self.send_keys(*keys_to_send)
        return self

    def send_keys_to(
        self, element: Element | WebElement, *keys_to_send: str
    ) -> _Actions:
        """Alias for send_keys_to_element"""
        return self.send_keys_to_element(element, *keys_to_send)

    def scroll_to_element(self, element: Element | WebElement) -> _Actions:
        """If the element is outside the viewport, scrolls the bottom of the
        element to the bottom of the viewport.

        :Args:
         - element: Which element to scroll into the viewport.
        """
        self._chain.scroll_to_element(_ensure_located(element))
        return self

    def scroll_to(self, element: Element | WebElement) -> _Actions:
        """Alias for scroll_to_element"""
        return self.scroll_to_element(element)

    def scroll_by_amount(self, delta_x: int, delta_y: int) -> _Actions:
        """Scrolls by provided amounts with the origin in the top left corner
        of the viewport.

        :Args:
         - delta_x: Distance along X axis to scroll using the wheel. A negative value scrolls left.
         - delta_y: Distance along Y axis to scroll using the wheel. A negative value scrolls up.
        """
        self._chain.scroll_by_amount(delta_x, delta_y)
        return self

    def scroll_by(self, delta_x: int, delta_y: int) -> _Actions:
        """Alias for scroll_by_amount"""
        return self.scroll_by_amount(delta_x, delta_y)

    def scroll_from_origin(
        self, scroll_origin: ScrollOrigin, delta_x: int, delta_y: int
    ) -> _Actions:
        """Scrolls by provided amount based on a provided origin. The scroll
        origin is either the center of an element or the upper left of the
        viewport plus any offsets. If the origin is an element, and the element
        is not in the viewport, the bottom of the element will first be
        scrolled to the bottom of the viewport.

        :Args:
         - origin: Where scroll originates (viewport or element center) plus provided offsets.
         - delta_x: Distance along X axis to scroll using the wheel. A negative value scrolls left.
         - delta_y: Distance along Y axis to scroll using the wheel. A negative value scrolls up.

         :Raises: If the origin with offset is outside the viewport.
          - MoveTargetOutOfBoundsException - If the origin with offset is outside the viewport.
        """
        self._chain.scroll_from_origin(scroll_origin, delta_x, delta_y)
        return self



# MIT License
#
# Copyright (c) 2015-2022 Iakiv Kramarenko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations
import sys
from typing import Union, Optional, overload

from selenium.webdriver import Keys
from selenium.webdriver.support import expected_conditions
from selenium.webdriver.support.wait import WebDriverWait

from selene.core.entity import Element, Collection
from selene.core._browser import Browser
from selene.core.exceptions import _SeleneError
from selene.core.wait import Command
from selenium.webdriver import ActionChains
from selenium.webdriver.common.actions import interaction
from selenium.webdriver.common.actions.action_builder import ActionBuilder
from selenium.webdriver.common.actions.pointer_input import PointerInput


# TODO: refactor to be of same style as __ClickWithOffset
#       in order to make autocomplete work properly
#       do it for save_screenshot and all other similar impls
def save_screenshot(path: Optional[str] = None) -> Command[Browser]:
    command: Command[Browser] = Command(
        'save screenshot',
        lambda browser: browser.config._save_screenshot_strategy(browser.config, path),
    )

    if isinstance(path, Browser):
        # somebody passed command as `.perform(command.save_screenshot)`
        # not as `.perform(command.save_screenshot())`
        browser = path
        command.__call__(browser)

    return command


def save_page_source(path: Optional[str] = None) -> Command[Browser]:
    command: Command[Browser] = Command(
        'save page source',
        lambda browser: browser.config._save_page_source_strategy(browser.config, path),
    )

    if isinstance(path, Browser):
        # somebody passed command as `.perform(command.save_screenshot)`
        # not as `.perform(command.save_screenshot())`
        browser = path
        command.__call__(browser)

    return command


def __select_all_actions(entity: Element | Browser):
    _COMMAND_KEY = Keys.COMMAND if sys.platform == 'darwin' else Keys.CONTROL
    actions: ActionChains = ActionChains(entity.config.driver)

    actions.key_down(_COMMAND_KEY)

    if isinstance(entity, Element):
        actions.send_keys_to_element(entity.locate(), 'a')
    else:
        actions.send_keys('a')

    actions.key_up(_COMMAND_KEY)

    actions.perform()


select_all: Command[Element | Browser] = Command(
    'send «select all» keys shortcut as ctrl+a or cmd+a for mac',
    __select_all_actions,
)


# TODO: can we make it work for both mobile and web?
#       should we selectively choose proper interaction.POINTER_TOUCH below?
def _long_press(duration=1.0):
    def func(element: Element):
        located_element = element.locate()
        driver = element.config.driver
        actions: ActionChains = ActionChains(driver)

        actions.w3c_actions = ActionBuilder(
            driver, mouse=PointerInput(interaction.POINTER_TOUCH, 'touch')
        )
        (
            actions.w3c_actions.pointer_action.move_to(located_element)
            .pointer_down()
            .pause(duration)
            .release()
        )
        actions.perform()

    command = Command(f'long press with duration={duration}', func)

    if isinstance(duration, Element):
        # somebody passed command as `.perform(command.long_press)`
        # not as `.perform(command.long_press())`
        # TODO: refactor to really allow such use case without conflicts on types
        element = duration
        command.__call__(element)

    return command


# TODO: consider
#       .with(ensure_state_changed=True).perform(command.drag_and_drop_to(target))
#       over
#       .perform(command.drag_and_drop_to(target, assert_location_changed=True))
#       but how should we interpret `ensure_state_changed` for other commands?
#       – case by case? i.e. location_changed for drag_and_drop_to
#         and something different for other commands?
#       – or maybe actually check the overall state of the page,
#         i.e. something like page_source_changed or dom_changed?
#       what about:
#       perform(command.drag_and_drop_to(target).to_change_location())
#       over
#       .perform(command.drag_and_drop_to(target, assert_location_changed=True))


def drag_and_drop_to(
    target: Element, _assert_location_changed: bool = False
) -> Command[Element]:
    """
    Args:
        target: a destination element to drag and drop to
        _assert_location_changed: False by default, but if True,
            then will assert that element was dragged to the new location,
            hence forcing a command retry if command was under waiting.
            This option is marked as experimental (by leading underscore),
            it may be renamed or removed completely.
    """

    def func(source: Element):
        source_webelement = source.locate()
        source_location = (
            source_webelement.location if _assert_location_changed else None
        )

        ActionChains(source.config.driver).drag_and_drop(
            source_webelement,
            target.locate(),
        ).perform()

        if _assert_location_changed and source_location == source.locate().location:
            raise _SeleneError('Element was not dragged to the new place')

    return Command(f'drag and drop to: {target}', func)


# TODO: consider adding 0 as default for x and y
def drag_and_drop_by_offset(x: int, y: int) -> Command[Element]:
    def func(source: Element):
        ActionChains(source.config.driver).drag_and_drop_by_offset(
            source.locate(),
            x,
            y,
        ).perform()

    return Command(f'drag and drop by offset: x={x}, y={y}', func)


class js:  # pylint: disable=invalid-name
    @staticmethod
    def set_value(value: Union[str, int]) -> Command[Element]:
        def func(element: Element):
            element.execute_script(
                """
                var text = arguments[0];
                var maxlength = element.getAttribute('maxlength') === null
                    ? -1
                    : parseInt(element.getAttribute('maxlength'));
                element.value = maxlength === -1
                    ? text
                    : text.length <= maxlength
                        ? text
                        : text.substring(0, maxlength);
                return null;
                """,
                str(value),
            )

        return Command(f'set value by js: {value}', func)

    @staticmethod
    def type(keys: Union[str, int]) -> Command[Element]:
        def func(element: Element):
            element.execute_script(
                """
                textToAppend = arguments[0];
                var value = element.value || '';
                var text = value + textToAppend;
                var maxlength = element.getAttribute('maxlength') === null
                    ? -1
                    : parseInt(element.getAttribute('maxlength'));
                element.value = maxlength === -1
                    ? text
                    : text.length <= maxlength
                        ? text
                        : text.substring(0, maxlength);
                return null;
                """,
                str(keys),
            )

        return Command(f'set value by js: {keys}', func)

    scroll_into_view: Command[Element] = Command(
        'scroll into view',
        lambda element: element.execute_script('element.scrollIntoView(true)'),
    )

    # TODO: should we process collections too? i.e. click through all elements?
    class __ClickWithOffset(Command[Element]):
        def __init__(self):
            self._description = 'click'

        @overload
        def __call__(self, element: Element) -> None: ...

        @overload
        def __call__(self, *, xoffset=0, yoffset=0) -> Command[Element]: ...

        def __call__(self, element: Element | None = None, *, xoffset=0, yoffset=0):
            def func(element: Element):
                element.execute_script(
                    '''
                    const offsetX = arguments[0]
                    const offsetY = arguments[1]
                    const rect = element.getBoundingClientRect()

                    function mouseEvent() {
                      if (typeof (Event) === 'function') {
                        return new MouseEvent('click', {
                          view: window,
                          bubbles: true,
                          cancelable: true,
                          clientX: rect.left + rect.width / 2 + offsetX,
                          clientY: rect.top + rect.height / 2 + offsetY
                        })
                      }
                      else {
                        const event = document.createEvent('MouseEvent')
                        event.initEvent('click', true, true)
                        event.type = 'click'
                        event.view = window
                        event.clientX = rect.left + rect.width / 2 + offsetX
                        event.clientY = rect.top + rect.height / 2 + offsetY
                        return event
                      }
                    }
                    element.dispatchEvent(mouseEvent())
                    ''',
                    xoffset,
                    yoffset,
                )

            if element is not None:
                # somebody passed command as `.perform(command.js.click)`
                # not as `.perform(command.js.click())`
                func(element)
                return None

            return Command(
                (
                    self.__str__()
                    if (not xoffset and not yoffset)
                    else f'click(xoffset={xoffset},yoffset={yoffset})'
                ),
                func,
            )

    click = __ClickWithOffset()

    clear_local_storage: Command[Browser] = Command(
        'clear local storage',
        lambda browser: browser.driver.execute_script('window.localStorage.clear()'),
    )

    clear_session_storage: Command[Browser] = Command(
        'clear local storage',
        lambda browser: browser.driver.execute_script('window.sessionStorage.clear()'),
    )

    remove: Command[Union[Element, Collection]] = Command(
        'remove',
        lambda entity: (
            entity.execute_script('element.remove()')
            if not hasattr(entity, '__iter__')
            else [element.execute_script('element.remove()') for element in entity]
        ),
    )

    @staticmethod
    def set_style_property(name: str, value: Union[str, int]) -> Command[Element]:
        return Command(
            f'set element.style.{name}="{value}"',
            lambda entity: (
                entity.execute_script(f'element.style.{name}="{value}"')
                if not hasattr(entity, '__iter__')
                else [
                    element.execute_script(f'element.style.{name}="{value}"')
                    for element in entity
                ]
            ),
        )

    set_style_display_to_none: Command[Union[Element, Collection]] = Command(
        'set element.style.display="none"',
        lambda entity: (
            entity.execute_script('element.style.display="none"')
            if not hasattr(entity, '__iter__')
            else [
                element.execute_script('element.style.display="none"')
                for element in entity
            ]
        ),
    )

    set_style_display_to_block: Command[Union[Element, Collection]] = Command(
        'set element.style.display="block"',
        lambda entity: (
            entity.execute_script('element.style.display="block"')
            if not hasattr(entity, '__iter__')
            else [
                element.execute_script('element.style.display="block"')
                for element in entity
            ]
        ),
    )

    set_style_visibility_to_hidden: Command[Union[Element, Collection]] = Command(
        'set element.style.visibility="hidden"',
        lambda entity: (
            entity.execute_script('element.style.visibility="hidden"')
            if not hasattr(entity, '__iter__')
            else [
                element.execute_script('element.style.visibility="hidden"')
                for element in entity
            ]
        ),
    )

    set_style_visibility_to_visible: Command[Union[Element, Collection]] = Command(
        'set element.style.visibility="visible"',
        lambda entity: (
            entity.execute_script('element.style.visibility="visible"')
            if not hasattr(entity, '__iter__')
            else [
                element.execute_script('element.style.visibility="visible"')
                for element in entity
            ]
        ),
    )

    # TODO: add js.drag_and_drop_by_offset(x, y)

    @staticmethod
    def drag_and_drop_to(target: Element) -> Command[Element]:
        """
        Simulates drag and drop via JavaScript.

        May not work everywhere. Among known cases:
        * does not work on https://mui.com/material-ui/react-slider/#ContinuousSlider
          where the normal drag and drop works fine.
        """

        def func(source: Element):
            script = """
            (function() {
              function createEvent(typeOfEvent) {
                var event = document.createEvent('CustomEvent');
                event.initCustomEvent(typeOfEvent, true, true, null);
                event.dataTransfer = {
                  data: {},
                  setData: function(key, value) {
                    this.data[key] = value;
                  },
                  getData: function(key) {
                    return this.data[key];
                  }
                };
                return event;
              }

              function dispatchEvent(element, event, transferData) {
                if (transferData !== undefined) {
                  event.dataTransfer = transferData;
                }
                if (element.dispatchEvent) {
                  element.dispatchEvent(event);
                } else if (element.fireEvent) {
                  element.fireEvent("on" + event.type, event);
                }
              }

              function dragAndDrop(element, target) {
                var dragStartEvent = createEvent('dragstart');
                dispatchEvent(element, dragStartEvent);
                var dropEvent = createEvent('drop');
                dispatchEvent(target, dropEvent, dragStartEvent.dataTransfer);
                var dragEndEvent = createEvent('dragend');
                dispatchEvent(element, dragEndEvent, dropEvent.dataTransfer);
              }

              return dragAndDrop(arguments[0], arguments[1]);
            })(...arguments)
            """.strip()
            source.config.driver.execute_script(
                script,
                source.locate(),
                target.locate(),
            )

        return Command(f'drag and drop to: {target}', func)

    @staticmethod
    def drop_file(path: str) -> Command[Element]:
        """
        Simulates via JavaScript the “drag and drop” of file into self (this element).

        The command is useful in cases,
        when there is no actual hidden input of type file to `send_keys(path)` to.

        Args:
            path: an absolute path to the file
        """

        # TODO: should we move them to params?
        #       what do they actually do? something like this? –
        #           xoffset: x offset (from this element center) to drop file
        #           yoffset: y offset (from this element center) to drop file
        xoffset = 0
        yoffset = 0

        def func(source: Element):
            script = """
            var target = arguments[0],
            offsetX = arguments[1],
            offsetY = arguments[2],
            document = target.ownerDocument || document,
            window = document.defaultView || window;

            var input = document.createElement('INPUT');
            input.type = 'file';
            input.style.display = 'none';
            input.onchange = function () {
              var rect = target.getBoundingClientRect(),
                  x = rect.left + (offsetX || (rect.width >> 1)),
                  y = rect.top + (offsetY || (rect.height >> 1)),
                  dataTransfer = {
                    files: this.files,
                    types: [ 'Files' ],
                  };

              ['dragenter', 'dragover', 'drop'].forEach(function (name) {
                var evt = document.createEvent('MouseEvent');
                evt.initMouseEvent(name, !0, !0, window, 0, 0, 0, x, y, !1, !1, !1, !1, 0, null);
                evt.dataTransfer = dataTransfer;
                target.dispatchEvent(evt);
              });

              setTimeout(function () { document.body.removeChild(input); }, 25);
            };
            document.body.appendChild(input);
            return input;
            """.strip()

            temp_input = source.config.driver.execute_script(
                script,
                source.locate(),
                xoffset,
                yoffset,
            )
            temp_input.send_keys(path)

            WebDriverWait(source.config.driver, 50).until(
                expected_conditions.staleness_of(temp_input)
            )

        return Command(f'drop file: {path}', func)
